@@ -35,7 +35,32 @@ struct _GMountSource
   char *obj_path;
 };
 
-G_DEFINE_TYPE (GMountSource, g_mount_source, G_TYPE_OBJECT);
+/* We use this hack to avoid problems when this code
+   is shared in both the daemon and the client */
+static GType _g_mount_source_get_type (void) G_GNUC_CONST;
+
+#define g_mount_source_get_type _g_mount_source_get_type
+G_DEFINE_TYPE (GMountSource, g_mount_source, G_TYPE_OBJECT)
+#undef g_mount_source_get_type
+
+GType
+g_mount_source_get_type (void)
+{
+  static volatile gsize type_volatile = 0;
+  
+  if (g_once_init_enter (&type_volatile))
+    {
+      GType type;
+      
+      type = g_type_from_name ("GMountSource");
+      if (type == 0)
+	type = _g_mount_source_get_type ();
+      
+      g_once_init_leave (&type_volatile, type);
+    }
+  
+  return type_volatile;
+}
 
 static void
 g_mount_source_finalize (GObject *object)
@@ -124,15 +149,17 @@ typedef struct AskPasswordData AskPasswordData;
 struct AskPasswordData {
 
   /* results: */
-  gboolean  aborted;
-  char     *password;
-  char     *username;
-  char     *domain;
+  gboolean       aborted;
+  char          *password;
+  char          *username;
+  char          *domain;
+  GPasswordSave  password_save;
+  gboolean	 anonymous;
 };
 
-typedef struct AskPasswordSyncData AskPasswordSyncData;
+typedef struct AskSyncData AskSyncData;
 
-struct AskPasswordSyncData {
+struct AskSyncData {
 
   /* For sync calls */
   GMutex *mutex;
@@ -192,9 +219,14 @@ ask_password_reply (DBusMessage *reply,
 	{
 	  data->aborted = aborted;
 
-	  data->password = g_strdup (password);
-	  data->username = g_strdup (username);
-	  data->domain = g_strdup (domain);
+	  if (!anonymous)
+	    {
+	      data->password = *password == 0 ? NULL : g_strdup (password);
+	      data->username = *username == 0 ? NULL : g_strdup (username);
+	      data->domain = *domain == 0 ? NULL : g_strdup (domain);
+	    }
+	  data->password_save = (GPasswordSave)password_save;
+	  data->anonymous = anonymous;
 
 	  /* TODO: handle more args */
 	}
@@ -202,7 +234,7 @@ ask_password_reply (DBusMessage *reply,
 
   if (handled == FALSE)
     {
-      g_simple_async_result_set_error (result, G_IO_ERROR, G_IO_ERROR_FAILED, NULL);
+      g_simple_async_result_set_error (result, G_IO_ERROR, G_IO_ERROR_FAILED, "Internal Error");
     }
 
   g_simple_async_result_complete (result);
@@ -228,7 +260,8 @@ g_mount_source_ask_password_async (GMountSource              *source,
       g_simple_async_report_error_in_idle (G_OBJECT (source),
 					   callback,
 					   user_data,
-					   G_IO_ERROR, G_IO_ERROR_FAILED, NULL); 
+					   G_IO_ERROR, G_IO_ERROR_FAILED, 
+					   "Internal Error"); 
       return;
     }
 
@@ -262,23 +295,46 @@ g_mount_source_ask_password_async (GMountSource              *source,
 
 }
 
+/**
+ * g_mount_source_ask_password_finish:
+ * @source: the source to query
+ * @result: the async result
+ * @aborted: set to %TRUE if the password dialog was aborted by the user
+ * @password_out: the to the password set by the user or to %NULL if none
+ * @user_out: set to the username set by the user or to %NULL if none
+ * @domain_out: set to the domain set by the user or to %NULL if none
+ * @anonymous_out: set to %TRUE if the user selected anonymous login. This
+ *                 should only happen if G_ASK_PASSWORD_ANONYMOUS_SUPPORTED
+ *                 was supplied whe querying the password.
+ * @password_save_out: set to the save flags to use when saving the password
+ *                     in the keyring.
+ *
+ * Requests the reply parameters from a g_mount_source_ask_password_async() 
+ * request. All out parameters can be set to %NULL to ignore them.
+ * <note><para>Please be aware that all string parameters can be set to %NULL,
+ * so make sure to check them.</para></note>
+ *
+ * Returns: %FALSE if the async reply contained an error.
+ **/
 gboolean
 g_mount_source_ask_password_finish (GMountSource  *source,
                                     GAsyncResult  *result,
                                     gboolean      *aborted,
                                     char         **password_out,
                                     char         **user_out,
-                                    char         **domain_out)
+                                    char         **domain_out,
+				    gboolean	  *anonymous_out,
+				    GPasswordSave *password_save_out)
 {
-  AskPasswordData *data;
+  AskPasswordData *data, def = { TRUE, };
   GSimpleAsyncResult *simple;
 
   simple = G_SIMPLE_ASYNC_RESULT (result);
 
   if (g_simple_async_result_propagate_error (simple, NULL))
-    return FALSE;
-
-  data = (AskPasswordData *) g_simple_async_result_get_op_res_gpointer (simple);
+    data = &def;
+  else
+    data = (AskPasswordData *) g_simple_async_result_get_op_res_gpointer (simple);
 
   if (aborted)
     *aborted = data->aborted;
@@ -301,21 +357,24 @@ g_mount_source_ask_password_finish (GMountSource  *source,
       data->domain = NULL;
     }
 
-  return TRUE;
+  if (anonymous_out)
+    *anonymous_out = data->anonymous;
+
+  if (password_save_out)
+    *password_save_out = data->password_save;  
+  
+  return data != &def;
 }
 
 
 static void
-ask_password_reply_sync  (GObject *source_object,
-                          GAsyncResult *res,
-                          gpointer user_data)
+ask_reply_sync  (GObject *source_object,
+		 GAsyncResult *res,
+		 gpointer user_data)
 {
-  GMountSource *source;
-  AskPasswordSyncData *data;
+  AskSyncData *data;
 
-  source = G_MOUNT_SOURCE (source_object);
-
-  data = (AskPasswordSyncData *) user_data;
+  data = (AskSyncData *) user_data;
 
   data->result = g_object_ref (res);
 
@@ -334,18 +393,12 @@ g_mount_source_ask_password (GMountSource *source,
 			     gboolean *aborted_out,
 			     char **password_out,
 			     char **user_out,
-			     char **domain_out)
+			     char **domain_out,
+			     gboolean *anonymous_out,
+			     GPasswordSave *password_save_out)
 {
-  char *password, *username, *domain;
-  gboolean handled, aborted;
-  AskPasswordSyncData data = {NULL};
-  
-  if (password_out)
-    *password_out = NULL;
-  if (user_out)
-    *user_out = NULL;
-  if (domain_out)
-    *domain_out = NULL;
+  gboolean handled;
+  AskSyncData data = {NULL};
   
   data.mutex = g_mutex_new ();
   data.cond = g_cond_new ();
@@ -358,7 +411,7 @@ g_mount_source_ask_password (GMountSource *source,
                                      default_user,
                                      default_domain,
                                      flags,
-                                     ask_password_reply_sync,
+                                     ask_reply_sync,
                                      &data);
   
   g_cond_wait(data.cond, data.mutex);
@@ -370,30 +423,14 @@ g_mount_source_ask_password (GMountSource *source,
 
   handled = g_mount_source_ask_password_finish (source,
                                                 data.result,
-                                                &aborted,
-                                                &password,
-                                                &username,
-                                                &domain);
+                                                aborted_out,
+                                                password_out,
+                                                user_out,
+                                                domain_out,
+						anonymous_out,
+						password_save_out);
   g_object_unref (data.result);
 
-  if (aborted_out)
-    *aborted_out = aborted;
-
-  if (password_out)
-    *password_out = password;
-  else
-    g_free (password);
-
-  if (user_out)
-    *user_out = username;
-  else
-    g_free (username);
-
-  if (domain_out)
-    *domain_out = domain;
-  else
-    g_free (domain);
-  
   return handled;
 }
 
@@ -409,6 +446,7 @@ op_ask_password_reply (GObject *source_object,
   char *username;
   char *password;
   char *domain;
+  GPasswordSave password_save;
 
   source = G_MOUNT_SOURCE (source_object);
   op = G_MOUNT_OPERATION (user_data);
@@ -421,7 +459,9 @@ op_ask_password_reply (GObject *source_object,
                                                 &aborted,
                                                 &username,
                                                 &password,
-                                                &domain);
+                                                &domain,
+						NULL,
+						&password_save);
 
   if (!handled)
     result = G_MOUNT_OPERATION_UNHANDLED;
@@ -437,6 +477,7 @@ op_ask_password_reply (GObject *source_object,
 	g_mount_operation_set_username (op, username);
       if (domain)
 	g_mount_operation_set_domain (op, domain);
+      g_mount_operation_set_password_save (op, password_save);
     }
   
   g_mount_operation_reply (op, result);
@@ -461,6 +502,225 @@ op_ask_password (GMountOperation *op,
   return TRUE;
 }
 
+typedef struct AskQuestionData AskQuestionData;
+
+struct AskQuestionData {
+
+  /* results: */
+  gboolean aborted;
+  guint32  choice;
+};
+
+/* the callback from dbus -> main thread */
+static void
+ask_question_reply (DBusMessage *reply,
+		    GError      *error,
+		    gpointer     _data)
+{
+  GSimpleAsyncResult *result;
+  AskQuestionData *data;
+  dbus_bool_t handled, aborted;
+  guint32 choice;
+  DBusMessageIter iter;
+
+  result = G_SIMPLE_ASYNC_RESULT (_data);
+  handled = TRUE;
+  
+  data = g_new0 (AskQuestionData, 1);
+  g_simple_async_result_set_op_res_gpointer (result, data, g_free);
+
+  if (reply == NULL)
+    {
+      data->aborted = TRUE;
+    }
+  else
+    {
+      dbus_message_iter_init (reply, &iter);
+      if (!_g_dbus_message_iter_get_args (&iter, NULL,
+					  DBUS_TYPE_BOOLEAN, &handled,
+					  DBUS_TYPE_BOOLEAN, &aborted,
+					  DBUS_TYPE_UINT32, &choice,
+					  0))
+	data->aborted = TRUE;
+      else
+	{
+	  data->aborted = aborted;
+	  data->choice = choice;
+	}
+    }
+
+  if (handled == FALSE)
+    {
+      g_simple_async_result_set_error (result, G_IO_ERROR, G_IO_ERROR_FAILED, "Internal Error");
+    }
+
+  g_simple_async_result_complete (result);
+}
+
+gboolean
+g_mount_source_ask_question (GMountSource *source,
+			     const char   *message,
+			     const char  **choices,
+			     gint          n_choices,
+			     gboolean     *aborted_out,
+			     gint         *choice_out)
+{
+  gint choice;
+  gboolean handled, aborted;
+  AskSyncData data = {NULL};
+  
+  data.mutex = g_mutex_new ();
+  data.cond = g_cond_new ();
+
+  g_mutex_lock (data.mutex);
+
+  g_mount_source_ask_question_async (source,
+                                     message,
+				     choices,
+				     n_choices,
+                                     ask_reply_sync,
+                                     &data);
+  
+  g_cond_wait(data.cond, data.mutex);
+  g_mutex_unlock (data.mutex);
+
+  g_cond_free (data.cond);
+  g_mutex_free (data.mutex);
+
+  handled = g_mount_source_ask_question_finish (source,
+                                                data.result,
+                                                &aborted,
+                                                &choice);
+  
+  g_object_unref (data.result);
+
+  if (aborted_out)
+    *aborted_out = aborted;
+
+  if (choice_out)
+    *choice_out = choice;
+  
+  return handled;	
+}
+
+void
+g_mount_source_ask_question_async (GMountSource       *source,
+				   const char         *message_string,
+				   const char        **choices,
+				   gint                n_choices,
+				   GAsyncReadyCallback callback,
+				   gpointer            user_data)
+{
+  GSimpleAsyncResult *result;
+  DBusMessage *message;
+
+  /* If no dbus id specified, reply that we weren't handled */
+  if (source->dbus_id[0] == 0)
+    { 
+      g_simple_async_report_error_in_idle (G_OBJECT (source),
+					   callback,
+					   user_data,
+					   G_IO_ERROR, G_IO_ERROR_FAILED, 
+					   "Internal Error"); 
+      return;
+    }
+
+  if (message_string == NULL)
+    message_string = "";
+  
+  message = dbus_message_new_method_call (source->dbus_id,
+					  source->obj_path,
+					  G_VFS_DBUS_MOUNT_OPERATION_INTERFACE,
+					  G_VFS_DBUS_MOUNT_OPERATION_OP_ASK_QUESTION);
+
+  _g_dbus_message_append_args (message,
+			       DBUS_TYPE_STRING, &message_string,
+			       DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
+			       choices, n_choices,
+			       0);
+
+  result = g_simple_async_result_new (G_OBJECT (source), callback, user_data, 
+                                      g_mount_source_ask_question_async);
+  /* 30 minute timeout */
+  _g_dbus_connection_call_async (NULL, message, 1000 * 60 * 30,
+				 ask_question_reply, result);
+  dbus_message_unref (message);
+	
+}
+
+gboolean
+g_mount_source_ask_question_finish (GMountSource *source,
+				    GAsyncResult *result,
+				    gboolean     *aborted,
+				    gint         *choice_out)
+{
+  AskQuestionData *data, def= { FALSE, };
+  GSimpleAsyncResult *simple;
+
+  simple = G_SIMPLE_ASYNC_RESULT (result);
+
+  if (g_simple_async_result_propagate_error (simple, NULL))
+    data = &def;
+  else
+    data = (AskQuestionData *) g_simple_async_result_get_op_res_gpointer (simple);
+
+  if (aborted)
+    *aborted = data->aborted;
+
+  if (choice_out)
+    *choice_out = data->choice;
+
+  return data != &def;	
+}
+
+static void
+op_ask_question_reply (GObject      *source_object,
+                       GAsyncResult *res,
+                       gpointer      user_data)
+{
+  GMountOperationResult result;
+  GMountOperation *op;
+  GMountSource *source;
+  gboolean handled, aborted;
+  gint choice;
+
+  source = G_MOUNT_SOURCE (source_object);
+  op = G_MOUNT_OPERATION (user_data);
+
+  handled = g_mount_source_ask_question_finish (source,
+                                                res,
+                                                &aborted,
+						&choice);
+
+  if (!handled)
+    result = G_MOUNT_OPERATION_UNHANDLED;
+  else if (aborted)
+    result = G_MOUNT_OPERATION_ABORTED;
+  else
+    {
+      result = G_MOUNT_OPERATION_HANDLED;
+      g_mount_operation_set_choice (op, choice);
+    }
+  
+  g_mount_operation_reply (op, result);
+  g_object_unref (op);
+}
+
+static gboolean
+op_ask_question (GMountOperation *op,
+		 const char      *message,
+		 const char     **choices,
+		 gint             n_choices,
+		 GMountSource    *mount_source)
+{
+  g_mount_source_ask_question_async (mount_source,
+				     message,
+				     choices,
+				     n_choices,
+				     op_ask_question_reply,
+				     g_object_ref (op));
+  return TRUE;
+}
 
 GMountOperation *
 g_mount_source_get_operation (GMountSource *mount_source)
@@ -474,6 +734,7 @@ g_mount_source_get_operation (GMountSource *mount_source)
 
 
   g_signal_connect (op, "ask_password", (GCallback)op_ask_password, mount_source);
+  g_signal_connect (op, "ask_question", (GCallback)op_ask_question, mount_source);
 
   return op;
 }
