@@ -188,10 +188,10 @@ auth_callback (SMBCCTX *context,
 	  in_keyring = g_vfs_keyring_lookup_password (backend->user,
 						      backend->server,
 						      backend->domain,
+						      "smb",
 						      NULL,
 						      NULL,
 						      0,
-						      "smb",
 						      &ask_user,
 						      &ask_domain,
 						      &ask_password);
@@ -222,6 +222,7 @@ auth_callback (SMBCCTX *context,
 						 &ask_password,
 						 &ask_user,
 						 &ask_domain,
+						 NULL,
 						 &(backend->password_save));
 	  g_free (message);
 	  if (!handled)
@@ -1843,3 +1844,9 @@ g_vfs_backend_smb_class_init (GVfsBackendSmbClass *klass)
 #endif
 
 }
+
+void
+g_vfs_smb_daemon_init (void)
+{
+  g_set_application_name (_("Windows Shares Filesystem Service"));
+}
