@@ -36,6 +36,9 @@
 #include "ghalmount.h"
 #include "ghalvolume.h"
 
+/* Protects all fields of GHalDrive that can change */
+G_LOCK_DEFINE_STATIC(hal_mount);
+
 struct _GHalMount {
   GObject parent;
 
@@ -71,6 +74,9 @@ _g_find_file_insensitive_async (GFile              *parent,
                                 GAsyncReadyCallback callback,
                                 gpointer            user_data);
 
+
+static GFile *get_root (GHalMount *hal_mount);
+
 static void g_hal_mount_mount_iface_init (GMountIface *iface);
 
 #define _G_IMPLEMENT_INTERFACE_DYNAMIC(TYPE_IFACE, iface_init)       { \
@@ -138,7 +144,20 @@ g_hal_mount_init (GHalMount *hal_mount)
 {
 }
 
-const static struct {
+static gboolean
+changed_in_idle (gpointer data)
+{
+  GHalMount *mount = data;
+  
+  g_signal_emit_by_name (mount, "changed");
+  if (mount->volume_monitor != NULL)
+    g_signal_emit_by_name (mount->volume_monitor, "mount_changed", mount);
+  g_object_unref (mount);
+  
+  return FALSE;
+}
+
+static const struct {
   const char *disc_type;
   const char *icon_name;
   char *ui_name;
@@ -161,7 +180,7 @@ const static struct {
   {"hddvd_r",       "media-optical-hddvd-r", N_("HD DVD-R Disc"), N_("Blank HD DVD-R Disc")},
   {"hddvd_rw",      "media-optical-hddvd-rw", N_("HD DVD-RW Disc"), N_("Blank HD DVD-RW Disc")},
   {"mo",            "media-optical-mo", N_("MO Disc"), N_("Blank MO Disc")},
-  {NULL,            "media-optical", N_("Disc"), N_("Blank Disc")},
+  {NULL,            "media-optical", N_("Disc"), N_("Blank Disc")}
 };
 
 static const char *
@@ -284,8 +303,9 @@ on_autorun_loaded (GObject *source_object, GAsyncResult *res,
       g_regex_unref (icon_regex);
       g_free (content);
     }
-  
-  if (relative_icon_path)
+
+  /* some autorun.in points to the .exe file for the icon; make sure we avoid using that */
+  if (relative_icon_path && !g_str_has_suffix (relative_icon_path, ".exe"))
     {
       _g_find_file_insensitive_async (data->root,
                                       relative_icon_path,
@@ -322,9 +342,9 @@ _g_find_mount_icon (GHalMount *m)
   
   m->searched_for_icon = TRUE;
 	
-  search_data = g_new (MountIconSearchData, 1);
+  search_data = g_new0 (MountIconSearchData, 1);
   search_data->mount = g_object_ref (m);
-  search_data->root = g_mount_get_root (G_MOUNT (m));
+  search_data->root = get_root (m);
   
   _g_find_file_insensitive_async (search_data->root,
                                   "autorun.inf",
@@ -366,10 +386,8 @@ do_update_from_hal (GHalMount *m)
 {
   HalDevice *volume;
   HalDevice *drive;
-
   char *name;
   const char *icon_name;
-
   const char *drive_type;
   const char *drive_bus;
   gboolean drive_uses_removable_media;
@@ -380,6 +398,13 @@ do_update_from_hal (GHalMount *m)
   gboolean volume_disc_has_data;
   const char *volume_disc_type;
   gboolean volume_disc_is_blank;
+  gboolean is_audio_player;
+  const char *icon_from_hal;
+  const char *volume_icon_from_hal;
+  const char *name_from_hal;
+  const char *volume_name_from_hal;
+  gboolean is_crypto;
+  gboolean is_crypto_cleartext;
 
   volume = m->device;
   drive = m->drive_device;
@@ -394,12 +419,33 @@ do_update_from_hal (GHalMount *m)
   volume_disc_has_data = hal_device_get_property_bool (volume, "volume.disc.has_data");
   volume_disc_is_blank = hal_device_get_property_bool (volume, "volume.disc.is_blank");
   volume_disc_type = hal_device_get_property_string (volume, "volume.disc.type");
-  
+  is_audio_player = hal_device_has_capability (drive, "portable_audio_player");
+  icon_from_hal = hal_device_get_property_string (drive, "info.desktop.icon");
+  volume_icon_from_hal = hal_device_get_property_string (volume, "info.desktop.icon");
+  name_from_hal = hal_device_get_property_string (drive, "info.desktop.name");
+  volume_name_from_hal = hal_device_get_property_string (volume, "info.desktop.name");
+
+  is_crypto = FALSE;
+  is_crypto_cleartext = FALSE;
+  if (strcmp (hal_device_get_property_string (volume, "volume.fsusage"), "crypto") == 0)
+      is_crypto = TRUE;
+
+  if (strlen (hal_device_get_property_string (volume, "volume.crypto_luks.clear.backing_volume")) > 0)
+      is_crypto_cleartext = TRUE;
+
   /*g_warning ("drive_type='%s'", drive_type); */
   /*g_warning ("drive_bus='%s'", drive_bus); */
   /*g_warning ("drive_uses_removable_media=%d", drive_uses_removable_media); */
-  
-  if (strcmp (drive_type, "disk") == 0)
+
+  if (strlen (volume_icon_from_hal) > 0)
+    icon_name = volume_icon_from_hal;
+  else if (strlen (icon_from_hal) > 0)
+    icon_name = icon_from_hal;
+  else if (is_audio_player)
+    icon_name = "multimedia-player";
+  else if (is_crypto || is_crypto_cleartext)
+    icon_name = "media-encrypted";
+  else if (strcmp (drive_type, "disk") == 0)
     {
       if (strcmp (drive_bus, "ide") == 0)
         icon_name = "drive-harddisk-ata";
@@ -430,7 +476,11 @@ do_update_from_hal (GHalMount *m)
     icon_name = "drive-harddisk";
 
   
-  if (volume_fs_label != NULL && strlen (volume_fs_label) > 0)
+  if (strlen (volume_name_from_hal) > 0)
+    name = g_strdup (volume_name_from_hal);
+  else if (strlen (name_from_hal) > 0)
+    name = g_strdup (name_from_hal);
+  else if (volume_fs_label != NULL && strlen (volume_fs_label) > 0)
     name = g_strdup (volume_fs_label);
   else if (volume_is_disc)
     {
@@ -474,6 +524,8 @@ update_from_hal (GHalMount *m, gboolean emit_changed)
   char *old_name;
   GIcon *old_icon;
 
+  G_LOCK (hal_mount);
+  
   old_name = g_strdup (m->name);
   old_icon = m->icon != NULL ? g_object_ref (m->icon) : NULL;
 
@@ -488,15 +540,13 @@ update_from_hal (GHalMount *m, gboolean emit_changed)
           old_icon == NULL ||
           strcmp (old_name, m->name) != 0 ||
           (! g_icon_equal (old_icon, m->icon)))
-        {
-          g_signal_emit_by_name (m, "changed");
-          if (m->volume_monitor != NULL)
-            g_signal_emit_by_name (m->volume_monitor, "mount_changed", m);
-        }
+        g_idle_add (changed_in_idle, g_object_ref (m));
     }
   g_free (old_name);
   if (old_icon != NULL)
     g_object_unref (old_icon);
+
+  G_UNLOCK (hal_mount);
 }
 
 static void
@@ -589,13 +639,16 @@ g_hal_mount_new_for_hal_device    (GVolumeMonitor    *volume_monitor,
 void 
 g_hal_mount_override_name (GHalMount *mount, const char *name)
 {
-  if (mount->override_name != NULL)
-    g_free (mount->override_name);
+  G_LOCK (hal_mount);
+  
+  g_free (mount->override_name);
 
   if (name != NULL)
     mount->override_name = g_strdup (name);
   else
     mount->override_name = NULL;
+  
+  G_UNLOCK (hal_mount);
 
   update_from_hal (mount, TRUE);
 }
@@ -603,6 +656,7 @@ g_hal_mount_override_name (GHalMount *mount, const char *name)
 void
 g_hal_mount_override_icon (GHalMount *mount, GIcon *icon)
 {
+  G_LOCK (hal_mount);
   if (mount->override_icon != NULL)
     g_object_unref (mount->override_icon);
 
@@ -611,6 +665,8 @@ g_hal_mount_override_icon (GHalMount *mount, GIcon *icon)
   else
     mount->override_icon = NULL;
 
+  G_UNLOCK (hal_mount);
+  
   update_from_hal (mount, TRUE);
 }
 
@@ -624,9 +680,9 @@ g_hal_mount_new (GVolumeMonitor       *volume_monitor,
   HalDevice *drive_device;
   const char *storage_udi;
   GHalMount *mount;
-  
-  /* No volume for mount: Ignore internal things */
-  if (volume == NULL && g_unix_mount_is_system_internal (mount_entry))
+
+  /* If no volume for mount - Ignore internal things */
+  if (volume == NULL && !g_unix_mount_guess_should_display (mount_entry))
     return NULL;
 
   mount = g_object_new (G_TYPE_HAL_MOUNT, NULL);
@@ -676,6 +732,12 @@ g_hal_mount_new (GVolumeMonitor       *volume_monitor,
 
  not_hal:
 
+  if (volume != NULL)
+    {
+      g_object_unref (mount);
+      return NULL;
+    }
+  
   mount->name = g_unix_mount_guess_name (mount_entry);
   mount->icon = g_unix_mount_guess_icon (mount_entry);
 
@@ -692,121 +754,175 @@ g_hal_mount_new (GVolumeMonitor       *volume_monitor,
 void
 g_hal_mount_unmounted (GHalMount *mount)
 {
+  G_LOCK (hal_mount);
   if (mount->volume != NULL)
     {
       g_hal_volume_unset_mount (mount->volume, mount);
       mount->volume = NULL;
-      g_signal_emit_by_name (mount, "changed");
-      /* there's really no need to emit volume_changed on the volume monitor 
-       * as we're going to be deleted.. */
+      g_idle_add (changed_in_idle, g_object_ref (mount));
     }
+  G_UNLOCK (hal_mount);
 }
 
 void
 g_hal_mount_unset_volume (GHalMount *mount,
                                        GHalVolume  *volume)
 {
+  G_LOCK (hal_mount);
   if (mount->volume == volume)
     {
       mount->volume = NULL;
-      /* TODO: Emit changed in idle to avoid locking issues */
-      g_signal_emit_by_name (mount, "changed");
-      if (mount->volume_monitor != NULL)
-        g_signal_emit_by_name (mount->volume_monitor, "mount_changed", mount);
+      g_idle_add (changed_in_idle, g_object_ref (mount));
     }
+  G_UNLOCK (hal_mount);
 }
 
 static GFile *
-g_hal_mount_get_root (GMount *mount)
+get_root (GHalMount *hal_mount)
 {
-  GHalMount *hal_mount = G_HAL_MOUNT (mount);
-
   if (hal_mount->override_root != NULL)
     return g_object_ref (hal_mount->override_root);
   else
     return g_file_new_for_path (hal_mount->mount_path);
 }
 
+static GFile *
+g_hal_mount_get_root (GMount *mount)
+{
+  GHalMount *hal_mount = G_HAL_MOUNT (mount);
+  GFile *root;
+
+  G_LOCK (hal_mount);
+  root = get_root (hal_mount);
+  G_UNLOCK (hal_mount);
+
+  return root;
+}
+
 static GIcon *
 g_hal_mount_get_icon (GMount *mount)
 {
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
+  GIcon *icon;
 
-  return g_object_ref (hal_mount->icon);
+  G_LOCK (hal_mount);
+  icon = g_object_ref (hal_mount->icon);
+  G_UNLOCK (hal_mount);
+  
+  return icon;
 }
 
 static char *
 g_hal_mount_get_uuid (GMount *mount)
 {
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
+  char *uuid;
 
-  return g_strdup (hal_mount->uuid);
+  G_LOCK (hal_mount);
+  uuid = g_strdup (hal_mount->uuid);
+  G_UNLOCK (hal_mount);
+  
+  return uuid;
 }
 
 static char *
 g_hal_mount_get_name (GMount *mount)
 {
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
+  char *name;
+
+  G_LOCK (hal_mount);
+  name = g_strdup (hal_mount->name);
+  G_UNLOCK (hal_mount);
   
-  return g_strdup (hal_mount->name);
+  return name;
 }
 
 gboolean
 g_hal_mount_has_uuid (GHalMount         *mount,
                        const char        *uuid)
 {
-  if (mount->uuid == NULL)
-    return FALSE;
-  return strcmp (mount->uuid, uuid) == 0;
+  gboolean res;
+
+  G_LOCK (hal_mount);
+  res = FALSE;
+  if (mount->uuid != NULL)
+    res = strcmp (mount->uuid, uuid) == 0;
+  G_UNLOCK (hal_mount);
+
+  return res;
 }
 
 gboolean
 g_hal_mount_has_mount_path (GHalMount *mount,
                             const char  *mount_path)
 {
-  return strcmp (mount->mount_path, mount_path) == 0;
+  gboolean res;
+
+  G_LOCK (hal_mount);
+  res = strcmp (mount->mount_path, mount_path) == 0;
+  G_UNLOCK (hal_mount);
+  return res;
 }
 
 gboolean
 g_hal_mount_has_udi (GHalMount *mount,
                      const char  *udi)
 {
-  if (mount->device == NULL)
-    return FALSE;
-  return strcmp (hal_device_get_udi (mount->device), udi) == 0;
+  gboolean res;
+
+  G_LOCK (hal_mount);
+  res = FALSE;
+  if (mount->device != NULL)
+    res = strcmp (hal_device_get_udi (mount->device), udi) == 0;
+  G_UNLOCK (hal_mount);
+
+  return res;
 }
 
 static GDrive *
 g_hal_mount_get_drive (GMount *mount)
 {
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
+  GDrive *drive;
 
+  G_LOCK (hal_mount);
+  drive = NULL;
   if (hal_mount->volume != NULL)
-    return g_volume_get_drive (G_VOLUME (hal_mount->volume));
+    drive = g_volume_get_drive (G_VOLUME (hal_mount->volume));
+  G_UNLOCK (hal_mount);
 
-  return NULL;
+  return drive;
 }
 
 static GVolume *
 g_hal_mount_get_volume (GMount *mount)
 {
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
+  GVolume *volume;
 
+  G_LOCK (hal_mount);
+  volume = NULL;
   if (hal_mount->volume)
-    return G_VOLUME (g_object_ref (hal_mount->volume));
+    volume = G_VOLUME (g_object_ref (hal_mount->volume));
+  G_UNLOCK (hal_mount);
   
-  return NULL;
+  return volume;
 }
 
 static gboolean
 g_hal_mount_can_unmount (GMount *mount)
 {
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
+  gboolean res;
 
+  G_LOCK (hal_mount);
+  res = TRUE;
   if (hal_mount->cannot_unmount)
-    return FALSE;
+    res = FALSE;
+  G_UNLOCK (hal_mount);
 
-  return TRUE;
+  return res;
 }
 
 static gboolean
@@ -816,19 +932,19 @@ g_hal_mount_can_eject (GMount *mount)
   GDrive *drive;
   gboolean can_eject;
 
+  G_LOCK (hal_mount);
   can_eject = FALSE;
-
   if (hal_mount->volume != NULL)
     {
       drive = g_volume_get_drive (G_VOLUME (hal_mount->volume));
       if (drive != NULL)
         can_eject = g_drive_can_eject (drive);
     }
+  G_UNLOCK (hal_mount);
 
   return can_eject;
 }
 
-
 typedef struct {
   GObject *object;
   GAsyncReadyCallback callback;
@@ -968,18 +1084,27 @@ g_hal_mount_unmount (GMount              *mount,
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
   char *argv[] = {"gnome-mount", "-u", "-b", "-d", NULL, NULL};
   gboolean using_legacy = FALSE;
+  char *d, *m;
 
+  G_LOCK (hal_mount);
+  d = g_strdup (hal_mount->device_path);
+  m = g_strdup (hal_mount->mount_path);
+  G_UNLOCK (hal_mount);
+
+  
   if (hal_mount->device != NULL)
-    argv[4] = hal_mount->device_path;
+    argv[4] = d;
   else
     {
       using_legacy = TRUE;
       argv[0] = "umount";
-      argv[1] = hal_mount->mount_path;
+      argv[1] = m;
       argv[2] = NULL;
     }
 
   unmount_do (mount, cancellable, callback, user_data, argv, using_legacy);
+  g_free (d);
+  g_free (m);
 }
 
 static gboolean
@@ -1016,18 +1141,21 @@ g_hal_mount_eject (GMount              *mount,
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
   GDrive *drive;
 
+  G_LOCK (hal_mount);
+  drive = NULL;
   if (hal_mount->volume != NULL)
+    drive = g_volume_get_drive (G_VOLUME (hal_mount->volume));
+  G_UNLOCK (hal_mount);
+  
+  if (drive != NULL)
     {
-      drive = g_volume_get_drive (G_VOLUME (hal_mount->volume));
-      if (drive != NULL)
-        {
-          EjectWrapperOp *data;
-          data = g_new0 (EjectWrapperOp, 1);
-          data->object = G_OBJECT (mount);
-          data->callback = callback;
-          data->user_data = user_data;
-          g_drive_eject (drive, flags, cancellable, eject_wrapper_callback, data);
-        }
+      EjectWrapperOp *data;
+      data = g_new0 (EjectWrapperOp, 1);
+      data->object = G_OBJECT (mount);
+      data->callback = callback;
+      data->user_data = user_data;
+      g_drive_eject (drive, flags, cancellable, eject_wrapper_callback, data);
+      g_object_unref (drive);
     }
 }
 
@@ -1038,14 +1166,22 @@ g_hal_mount_eject_finish (GMount        *mount,
 {
   GHalMount *hal_mount = G_HAL_MOUNT (mount);
   GDrive *drive;
+  gboolean res;
 
+  res = TRUE;
+  
+  G_LOCK (hal_mount);
+  drive = NULL;
   if (hal_mount->volume != NULL)
+    drive = g_volume_get_drive (G_VOLUME (hal_mount->volume));
+  G_UNLOCK (hal_mount);
+  
+  if (drive != NULL)
     {
-      drive = g_volume_get_drive (G_VOLUME (hal_mount->volume));
-      if (drive != NULL)
-        return g_drive_eject_finish (drive, result, error);
+      res = g_drive_eject_finish (drive, result, error);
+      g_object_unref (drive);
     }
-  return TRUE;
+  return res;
 }
 
 static void
@@ -1081,6 +1217,11 @@ static void
 more_files_callback (GObject *source_object, GAsyncResult *res,
                      gpointer user_data);
 
+static void
+find_file_insensitive_exists_callback (GObject *source_object,
+                                       GAsyncResult *res,
+                                       gpointer user_data);
+
 typedef struct _InsensitiveFileSearchData
 {
 	GFile *root;
@@ -1103,29 +1244,21 @@ _g_find_file_insensitive_async (GFile              *parent,
                                 gpointer            user_data)
 {
   InsensitiveFileSearchData *data;
+  GFile *direct_file = g_file_get_child (parent, name);
   
-  g_return_if_fail (g_utf8_validate (name, -1, NULL));
-
-  data = g_new (InsensitiveFileSearchData, 1);
-  data->root = g_object_ref (parent);
-  data->original_path = g_strdup (name);
-  data->split_path = g_strsplit (name, G_DIR_SEPARATOR_S, -1);
-  data->index = 0;
-  data->enumerator = NULL;
-  data->current_file = g_object_ref (parent);
+  data = g_new0 (InsensitiveFileSearchData, 1);
   data->cancellable = cancellable;
   data->callback = callback;
   data->user_data = user_data;
-
-  /* Skip any empty components due to multiple slashes */
-  while (data->split_path[data->index] != NULL &&
-         *data->split_path[data->index] == 0)
-    data->index++;
+  data->root = g_object_ref (parent);
+  data->original_path = g_strdup (name);
+  
+  g_file_query_info_async (direct_file, G_FILE_ATTRIBUTE_STANDARD_TYPE,
+			   G_FILE_QUERY_INFO_NONE, G_PRIORITY_DEFAULT,
+			   cancellable,
+			   find_file_insensitive_exists_callback, data);
+  
   
-  g_file_enumerate_children_async (data->current_file,
-                                   G_FILE_ATTRIBUTE_STANDARD_NAME,
-                                   0, G_PRIORITY_DEFAULT, cancellable,
-                                   enumerated_children_callback, data);
 }
 
 static void
@@ -1133,8 +1266,7 @@ clear_find_file_insensitive_state (InsensitiveFileSearchData *data)
 {
   if (data->root)
     g_object_unref (data->root);
-  if (data->original_path)
-    g_free (data->original_path);
+  g_free (data->original_path);
   if (data->split_path)
     g_strfreev (data->split_path);
   if (data->enumerator)
@@ -1144,6 +1276,52 @@ clear_find_file_insensitive_state (InsensitiveFileSearchData *data)
   g_free (data);
 }
 
+static void
+find_file_insensitive_exists_callback (GObject *source_object,
+                                       GAsyncResult *res,
+                                       gpointer user_data)
+{
+  GFileInfo *info;
+  InsensitiveFileSearchData *data = (InsensitiveFileSearchData *) (user_data);
+  
+  /* The file exists and can be found with the given path, no need to search. */
+  if ((info = g_file_query_info_finish (G_FILE (source_object), res, NULL)))
+    {
+      GSimpleAsyncResult *simple;
+      
+      simple = g_simple_async_result_new (G_OBJECT (data->root),
+                                          data->callback,
+                                          data->user_data,
+                                          _g_find_file_insensitive_async);
+      
+      g_simple_async_result_set_op_res_gpointer (simple, g_object_ref (source_object), g_object_unref);
+      g_simple_async_result_complete_in_idle (simple);
+      g_object_unref (simple);
+      clear_find_file_insensitive_state (data);
+    }
+  
+  else
+    {
+      data->split_path = g_strsplit (data->original_path, G_DIR_SEPARATOR_S, -1);
+      data->index = 0;
+      data->enumerator = NULL;
+      data->current_file = g_object_ref (data->root);
+      
+      /* Skip any empty components due to multiple slashes */
+      while (data->split_path[data->index] != NULL &&
+             *data->split_path[data->index] == 0)
+        data->index++;
+      
+      g_file_enumerate_children_async (data->current_file,
+                                       G_FILE_ATTRIBUTE_STANDARD_NAME,
+                                       0, G_PRIORITY_DEFAULT,
+                                       data->cancellable,
+                                       enumerated_children_callback, data);
+    }
+  
+  g_object_unref (source_object);
+}
+
 static void
 enumerated_children_callback (GObject *source_object, GAsyncResult *res,
                               gpointer user_data)
@@ -1190,7 +1368,7 @@ more_files_callback (GObject *source_object, GAsyncResult *res,
   GList *files, *l;
   gchar *filename = NULL, *component, *case_folded_name,
     *name_collation_key;
-  gboolean end_of_files;
+  gboolean end_of_files, is_utf8;
   
   files = g_file_enumerator_next_files_finish (data->enumerator,
                                                res, NULL);
@@ -1200,29 +1378,44 @@ more_files_callback (GObject *source_object, GAsyncResult *res,
   component = data->split_path[data->index];
   g_return_if_fail (component != NULL);
   
-  case_folded_name = g_utf8_casefold (component, -1);
-  name_collation_key = g_utf8_collate_key (case_folded_name, -1);
-  g_free (case_folded_name);
+  is_utf8 = (g_utf8_validate (component, -1, NULL));
+  if (is_utf8)
+    {
+      case_folded_name = g_utf8_casefold (component, -1);
+      name_collation_key = g_utf8_collate_key (case_folded_name, -1);
+      g_free (case_folded_name);
+    }
+  
+  else
+    {
+      name_collation_key = g_ascii_strdown (component, -1);
+    }
   
   for (l = files; l != NULL; l = l->next)
     {
       GFileInfo *info;
       const gchar *this_name;
+      gchar *key;
       
       info = l->data;
       this_name = g_file_info_get_name (info);
       
-      if (g_utf8_validate (this_name, -1, NULL))
+      if (is_utf8 && g_utf8_validate (this_name, -1, NULL))
         {
-          gchar *case_folded, *key;
+          gchar *case_folded;
           case_folded = g_utf8_casefold (this_name, -1);
           key = g_utf8_collate_key (case_folded, -1);
           g_free (case_folded);
-          
-          if (strcmp (key, name_collation_key) == 0)
-            filename = g_strdup (this_name);
-          g_free (key);
         }
+      else
+        {
+          key = g_ascii_strdown (this_name, -1);
+        }
+      
+      if (strcmp (key, name_collation_key) == 0)
+          filename = g_strdup (this_name);
+      g_free (key);
+      
       if (filename)
         break;
     }
@@ -1235,6 +1428,10 @@ more_files_callback (GObject *source_object, GAsyncResult *res,
     {
       GFile *next_file;
       
+      g_file_enumerator_close_async (data->enumerator,
+                                     G_PRIORITY_DEFAULT,
+                                     data->cancellable,
+                                     NULL, NULL);
       g_object_unref (data->enumerator);
       data->enumerator = NULL;
       
