@@ -59,7 +59,6 @@ prompt_for (const char *prompt, const char *default_value, gboolean echo)
 #ifdef HAVE_TERMIOS_H
   struct termios term_attr; 
   int old_flags;
-  int res;
   gboolean restore_flags;
 #endif
   char data[256];
@@ -74,7 +73,7 @@ prompt_for (const char *prompt, const char *default_value, gboolean echo)
 
 #ifdef HAVE_TERMIOS_H
   restore_flags = FALSE;
-  if (!echo && (res = tcgetattr (STDIN_FILENO, &term_attr)) == 0)
+  if (!echo && tcgetattr (STDIN_FILENO, &term_attr) == 0)
     {
       old_flags = term_attr.c_lflag; 
       term_attr.c_lflag &= ~ECHO;
@@ -99,6 +98,9 @@ prompt_for (const char *prompt, const char *default_value, gboolean echo)
   len = strlen (data);
   if (len > 0 && data[len-1] == '\n')
     data[len-1] = 0;
+
+  if (!echo)
+    g_print ("\n");
   
   if (*data == 0 && default_value)
     return g_strdup (default_value);
@@ -203,9 +205,9 @@ mount (GFile *file)
   op = new_mount_op ();
 
   if (mount_mountable)
-    g_file_mount_mountable (file, op, NULL, mount_mountable_done_cb, op);
+    g_file_mount_mountable (file, 0, op, NULL, mount_mountable_done_cb, op);
   else
-    g_file_mount_enclosing_volume (file, op, NULL, mount_done_cb, op);
+    g_file_mount_enclosing_volume (file, 0, op, NULL, mount_done_cb, op);
     
   outstanding_mounts++;
 }
@@ -325,12 +327,13 @@ list_volumes (GList *volumes,
 	      gboolean only_with_no_drive)
 {
   GList *l, *mounts;
-  int c;
+  int c, i;
   GMount *mount;
   GVolume *volume;
   GDrive *drive;
   char *name;
   char *uuid;
+  char **ids;
   
   for (c = 0, l = volumes; l != NULL; l = l->next, c++)
     {
@@ -353,6 +356,20 @@ list_volumes (GList *volumes,
 
       if (mount_list_info)
 	{
+	  ids = g_volume_enumerate_identifiers (volume);
+	  if (ids && ids[0] != NULL)
+	    {
+	      g_print ("%*sids:\n", indent+2, "");
+	      for (i = 0; ids[i] != NULL; i++)
+		{
+		  char *id = g_volume_get_identifier (volume,
+						      ids[i]);
+		  g_print ("%*s %s: '%s'\n", indent+2, "", ids[i], id);
+		  g_free (id);
+		}
+	    }
+	  g_strfreev (ids);
+	  
 	  uuid = g_volume_get_uuid (volume);
 	  if (uuid)
 	    g_print ("%*suuid=%s\n", indent + 2, "", uuid);
@@ -377,9 +394,10 @@ list_drives (GList *drives,
 	     int indent)
 {
   GList *volumes, *l;
-  int c;
+  int c, i;
   GDrive *drive;
   char *name;
+  char **ids;
   
   for (c = 0, l = drives; l != NULL; l = l->next, c++)
     {
@@ -388,9 +406,23 @@ list_drives (GList *drives,
       
       g_print ("%*sDrive(%d): %s\n", indent, "", c, name);
       g_free (name);
-
+      
       if (mount_list_info)
 	{
+	  ids = g_drive_enumerate_identifiers (drive);
+	  if (ids && ids[0] != NULL)
+	    {
+	      g_print ("%*sids:\n", indent+2, "");
+	      for (i = 0; ids[i] != NULL; i++)
+		{
+		  char *id = g_drive_get_identifier (drive,
+						     ids[i]);
+		  g_print ("%*s %s: '%s'\n", indent+2, "", ids[i], id);
+		  g_free (id);
+		}
+	    }
+	  g_strfreev (ids);
+
 	  g_print ("%*sis_media_removable=%d\n", indent + 2, "", g_drive_is_media_removable (drive));
 	  g_print ("%*shas_media=%d\n", indent + 2, "", g_drive_has_media (drive));
 	  g_print ("%*sis_media_check_automatic=%d\n", indent + 2, "", g_drive_is_media_check_automatic (drive));
